@@ -12,14 +12,14 @@ NTL_START_IMPL
 
 const ZZ_pEX& ZZ_pEX::zero()
 {
-   static ZZ_pEX z;
+   NTL_THREAD_LOCAL static ZZ_pEX z;
    return z;
 }
 
 
 istream& operator>>(istream& s, ZZ_pEX& x)
 {
-   s >> x.rep;
+   NTL_INPUT_CHECK_RET(s, s >> x.rep);
    x.normalize();
    return s;
 }
@@ -111,10 +111,10 @@ void SetCoeff(ZZ_pEX& x, long i, const ZZ_pE& a)
    long j, m;
 
    if (i < 0) 
-      Error("SetCoeff: negative index");
+      LogicError("SetCoeff: negative index");
 
    if (NTL_OVERFLOW(i, 1, 0))
-      Error("overflow in SetCoeff");
+      ResourceError("overflow in SetCoeff");
 
    m = deg(x);
 
@@ -150,10 +150,10 @@ void SetCoeff(ZZ_pEX& x, long i, const ZZ_p& aa)
    long j, m;
 
    if (i < 0)
-      Error("SetCoeff: negative index");
+      LogicError("SetCoeff: negative index");
 
    if (NTL_OVERFLOW(i, 1, 0))
-      Error("overflow in SetCoeff");
+      ResourceError("overflow in SetCoeff");
 
    NTL_ZZ_pRegister(a);  // watch out for aliases!
    a = aa;
@@ -189,10 +189,10 @@ void SetCoeff(ZZ_pEX& x, long i)
    long j, m;
 
    if (i < 0) 
-      Error("coefficient index out of range");
+      LogicError("coefficient index out of range");
 
    if (NTL_OVERFLOW(i, 1, 0))
-      Error("overflow in SetCoeff");
+      ResourceError("overflow in SetCoeff");
 
    m = deg(x);
 
@@ -609,7 +609,7 @@ void mul(ZZ_pEX& c, const ZZ_pEX& a, const ZZ_pEX& b)
    long n2 = 2*n-1;
 
    if (NTL_OVERFLOW(da+db+1, n2, 0))
-      Error("overflow in ZZ_pEX mul");
+      ResourceError("overflow in ZZ_pEX mul");
 
    long i, j;
 
@@ -745,7 +745,7 @@ void sqr(ZZ_pEX& c, const ZZ_pEX& a)
    long n2 = 2*n-1;
 
    if (NTL_OVERFLOW(2*da+1, n2, 0))
-      Error("overflow in ZZ_pEX sqr");
+      ResourceError("overflow in ZZ_pEX sqr");
 
    long i, j;
 
@@ -787,7 +787,7 @@ void sqr(ZZ_pEX& c, const ZZ_pEX& a)
 
 void MulTrunc(ZZ_pEX& x, const ZZ_pEX& a, const ZZ_pEX& b, long n)
 {
-   if (n < 0) Error("MulTrunc: bad args");
+   if (n < 0) LogicError("MulTrunc: bad args");
 
    ZZ_pEX t;
    mul(t, a, b);
@@ -796,7 +796,7 @@ void MulTrunc(ZZ_pEX& x, const ZZ_pEX& a, const ZZ_pEX& b, long n)
 
 void SqrTrunc(ZZ_pEX& x, const ZZ_pEX& a, long n)
 {
-   if (n < 0) Error("SqrTrunc: bad args");
+   if (n < 0) LogicError("SqrTrunc: bad args");
 
    ZZ_pEX t;
    sqr(t, a);
@@ -837,7 +837,7 @@ void trunc(ZZ_pEX& x, const ZZ_pEX& a, long m)
 // x = a % X^m, output may alias input 
 
 {
-   if (m < 0) Error("trunc: bad args");
+   if (m < 0) LogicError("trunc: bad args");
 
    if (&x == &a) {
       if (x.rep.length() > m) {
@@ -903,7 +903,7 @@ void MulByXModAux(ZZ_pEX& h, const ZZ_pEX& a, const ZZ_pEX& f)
    n = deg(f);
    m = deg(a);
 
-   if (m >= n || n == 0) Error("MulByXMod: bad args");
+   if (m >= n || n == 0) LogicError("MulByXMod: bad args");
 
    if (m < 0) {
       clear(h);
@@ -986,15 +986,15 @@ void PlainMul(ZZ_pEX& x, const ZZ_pEX& a, const ZZ_pEX& b)
    xp = x.rep.elts();
 
    long i, j, jmin, jmax;
-   static ZZ_pX t, accum;
+   ZZ_pX t, accum;
 
    for (i = 0; i <= d; i++) {
       jmin = max(0, i-db);
       jmax = min(da, i);
       clear(accum);
       for (j = jmin; j <= jmax; j++) {
-	 mul(t, rep(ap[j]), rep(bp[i-j]));
-	 add(accum, accum, t);
+         mul(t, rep(ap[j]), rep(bp[i-j]));
+         add(accum, accum, t);
       }
       conv(xp[i], accum);
    }
@@ -1025,7 +1025,7 @@ void PlainDivRem(ZZ_pEX& q, ZZ_pEX& r, const ZZ_pEX& a, const ZZ_pEX& b)
    da = deg(a);
    db = deg(b);
 
-   if (db < 0) Error("ZZ_pEX: division by zero");
+   if (db < 0) ArithmeticError("ZZ_pEX: division by zero");
 
    if (da < db) {
       r = a;
@@ -1065,13 +1065,13 @@ void PlainDivRem(ZZ_pEX& q, ZZ_pEX& r, const ZZ_pEX& a, const ZZ_pEX& b)
    for (i = dq; i >= 0; i--) {
       conv(t, xp[i+db]);
       if (!LCIsOne)
-	 mul(t, t, LCInv);
+         mul(t, t, LCInv);
       qp[i] = t;
       negate(t, t);
 
       for (j = db-1; j >= 0; j--) {
-	 mul(s, rep(t), rep(bp[j]));
-	 add(xp[i+j], xp[i+j], s);
+         mul(s, rep(t), rep(bp[j]));
+         add(xp[i+j], xp[i+j], s);
       }
    }
 
@@ -1095,7 +1095,7 @@ void PlainRem(ZZ_pEX& r, const ZZ_pEX& a, const ZZ_pEX& b, vec_ZZ_pX& x)
    da = deg(a);
    db = deg(b);
 
-   if (db < 0) Error("ZZ_pEX: division by zero");
+   if (db < 0) ArithmeticError("ZZ_pEX: division by zero");
 
    if (da < db) {
       r = a;
@@ -1121,12 +1121,12 @@ void PlainRem(ZZ_pEX& r, const ZZ_pEX& a, const ZZ_pEX& b, vec_ZZ_pX& x)
    for (i = dq; i >= 0; i--) {
       conv(t, xp[i+db]);
       if (!LCIsOne)
-	 mul(t, t, LCInv);
+         mul(t, t, LCInv);
       negate(t, t);
 
       for (j = db-1; j >= 0; j--) {
-	 mul(s, rep(t), rep(bp[j]));
-	 add(xp[i+j], xp[i+j], s);
+         mul(s, rep(t), rep(bp[j]));
+         add(xp[i+j], xp[i+j], s);
       }
    }
 
@@ -1152,7 +1152,7 @@ void PlainDivRem(ZZ_pEX& q, ZZ_pEX& r, const ZZ_pEX& a, const ZZ_pEX& b,
    da = deg(a);
    db = deg(b);
 
-   if (db < 0) Error("ZZ_pEX: division by zero");
+   if (db < 0) ArithmeticError("ZZ_pEX: division by zero");
 
    if (da < db) {
       r = a;
@@ -1188,13 +1188,13 @@ void PlainDivRem(ZZ_pEX& q, ZZ_pEX& r, const ZZ_pEX& a, const ZZ_pEX& b,
    for (i = dq; i >= 0; i--) {
       conv(t, xp[i+db]);
       if (!LCIsOne)
-	 mul(t, t, LCInv);
+         mul(t, t, LCInv);
       qp[i] = t;
       negate(t, t);
 
       for (j = db-1; j >= 0; j--) {
-	 mul(s, rep(t), rep(bp[j]));
-	 add(xp[i+j], xp[i+j], s);
+         mul(s, rep(t), rep(bp[j]));
+         add(xp[i+j], xp[i+j], s);
       }
    }
 
@@ -1219,7 +1219,7 @@ void PlainDiv(ZZ_pEX& q, const ZZ_pEX& a, const ZZ_pEX& b)
    da = deg(a);
    db = deg(b);
 
-   if (db < 0) Error("ZZ_pEX: division by zero");
+   if (db < 0) ArithmeticError("ZZ_pEX: division by zero");
 
    if (da < db) {
       clear(q);
@@ -1257,15 +1257,15 @@ void PlainDiv(ZZ_pEX& q, const ZZ_pEX& a, const ZZ_pEX& b)
    for (i = dq; i >= 0; i--) {
       conv(t, xp[i]);
       if (!LCIsOne)
-	 mul(t, t, LCInv);
+         mul(t, t, LCInv);
       qp[i] = t;
       negate(t, t);
 
       long lastj = max(0, db-i);
 
       for (j = db-1; j >= lastj; j--) {
-	 mul(s, rep(t), rep(bp[j]));
-	 add(xp[i+j-db], xp[i+j-db], s);
+         mul(s, rep(t), rep(bp[j]));
+         add(xp[i+j-db], xp[i+j-db], s);
       }
    }
 }
@@ -1283,7 +1283,7 @@ void PlainRem(ZZ_pEX& r, const ZZ_pEX& a, const ZZ_pEX& b)
    da = deg(a);
    db = deg(b);
 
-   if (db < 0) Error("ZZ_pEX: division by zero");
+   if (db < 0) ArithmeticError("ZZ_pEX: division by zero");
 
    if (da < db) {
       r = a;
@@ -1312,12 +1312,12 @@ void PlainRem(ZZ_pEX& r, const ZZ_pEX& a, const ZZ_pEX& b)
    for (i = dq; i >= 0; i--) {
       conv(t, xp[i+db]);
       if (!LCIsOne)
-	 mul(t, t, LCInv);
+         mul(t, t, LCInv);
       negate(t, t);
 
       for (j = db-1; j >= 0; j--) {
-	 mul(s, rep(t), rep(bp[j]));
-	 add(xp[i+j], xp[i+j], s);
+         mul(s, rep(t), rep(bp[j]));
+         add(xp[i+j], xp[i+j], s);
       }
    }
 
@@ -1337,7 +1337,7 @@ void RightShift(ZZ_pEX& x, const ZZ_pEX& a, long n)
    }
 
    if (n < 0) {
-      if (n < -NTL_MAX_LONG) Error("overflow in RightShift");
+      if (n < -NTL_MAX_LONG) ResourceError("overflow in RightShift");
       LeftShift(x, a, -n);
       return;
    }
@@ -1378,7 +1378,7 @@ void LeftShift(ZZ_pEX& x, const ZZ_pEX& a, long n)
    }
 
    if (NTL_OVERFLOW(n, 1, 0))
-      Error("overflow in LeftShift");
+      ResourceError("overflow in LeftShift");
 
    long m = a.rep.length();
 
@@ -1405,7 +1405,7 @@ void NewtonInv(ZZ_pEX& c, const ZZ_pEX& a, long e)
       return;
    }
 
-   static vec_long E;
+   vec_long E;
    E.SetLength(0);
    append(E, e);
    while (e > 1) {
@@ -1451,14 +1451,14 @@ void NewtonInv(ZZ_pEX& c, const ZZ_pEX& a, long e)
 
 void InvTrunc(ZZ_pEX& c, const ZZ_pEX& a, long e)
 {
-   if (e < 0) Error("InvTrunc: bad args");
+   if (e < 0) LogicError("InvTrunc: bad args");
    if (e == 0) {
       clear(c);
       return;
    }
 
    if (NTL_OVERFLOW(e, 1, 0))
-      Error("overflow in InvTrunc");
+      ResourceError("overflow in InvTrunc");
 
    NewtonInv(c, a, e);
 }
@@ -1474,13 +1474,13 @@ void build(ZZ_pEXModulus& F, const ZZ_pEX& f)
 {
    long n = deg(f);
 
-   if (n <= 0) Error("build(ZZ_pEXModulus,ZZ_pEX): deg(f) <= 0");
+   if (n <= 0) LogicError("build(ZZ_pEXModulus,ZZ_pEX): deg(f) <= 0");
 
    if (NTL_OVERFLOW(n, ZZ_pE::degree(), 0))
-      Error("build(ZZ_pEXModulus,ZZ_pEX): overflow");
+      ResourceError("build(ZZ_pEXModulus,ZZ_pEX): overflow");
    
 
-   F.tracevec.SetLength(0);
+   F.tracevec.make();
 
    F.f = f;
    F.n = n;
@@ -1736,7 +1736,7 @@ void div(ZZ_pEX& q, const ZZ_pEX& a, const ZZ_pEXModulus& F)
 
 void MulMod(ZZ_pEX& c, const ZZ_pEX& a, const ZZ_pEX& b, const ZZ_pEXModulus& F)
 {
-   if (deg(a) >= F.n || deg(b) >= F.n) Error("MulMod: bad args");
+   if (deg(a) >= F.n || deg(b) >= F.n) LogicError("MulMod: bad args");
 
    ZZ_pEX t;
    mul(t, a, b);
@@ -1746,7 +1746,7 @@ void MulMod(ZZ_pEX& c, const ZZ_pEX& a, const ZZ_pEX& b, const ZZ_pEXModulus& F)
 
 void SqrMod(ZZ_pEX& c, const ZZ_pEX& a, const ZZ_pEXModulus& F)
 {
-   if (deg(a) >= F.n) Error("MulMod: bad args");
+   if (deg(a) >= F.n) LogicError("MulMod: bad args");
 
    ZZ_pEX t;
    sqr(t, a);
@@ -2057,7 +2057,7 @@ void eval(vec_ZZ_pE& b, const ZZ_pEX& f, const vec_ZZ_pE& a)
 void interpolate(ZZ_pEX& f, const vec_ZZ_pE& a, const vec_ZZ_pE& b)
 {
    long m = a.length();
-   if (b.length() != m) Error("interpolate: vector length mismatch");
+   if (b.length() != m) LogicError("interpolate: vector length mismatch");
 
    if (m == 0) {
       clear(f);
@@ -2185,7 +2185,7 @@ void build(ZZ_pEXArgument& A, const ZZ_pEX& h, const ZZ_pEXModulus& F, long m)
    long i;
 
    if (m <= 0 || deg(h) >= F.n)
-      Error("build: bad args");
+      LogicError("build: bad args");
 
    if (m > F.n) m = F.n;
 
@@ -2209,7 +2209,7 @@ void build(ZZ_pEXArgument& A, const ZZ_pEX& h, const ZZ_pEXModulus& F, long m)
       MulMod(A.H[i], A.H[i-1], h, F);
 }
 
-long ZZ_pEXArgBound = 0;
+NTL_THREAD_LOCAL long ZZ_pEXArgBound = 0;
 
 
 
@@ -2292,7 +2292,7 @@ void build(ZZ_pEXTransMultiplier& B, const ZZ_pEX& b, const ZZ_pEXModulus& F)
 {
    long db = deg(b);
 
-   if (db >= F.n) Error("build TransMultiplier: bad args");
+   if (db >= F.n) LogicError("build TransMultiplier: bad args");
 
    ZZ_pEX t;
 
@@ -2334,7 +2334,7 @@ void build(ZZ_pEXTransMultiplier& B, const ZZ_pEX& b, const ZZ_pEXModulus& F)
 void TransMulMod(ZZ_pEX& x, const ZZ_pEX& a, const ZZ_pEXTransMultiplier& B,
                const ZZ_pEXModulus& F)
 {
-   if (deg(a) >= F.n) Error("TransMulMod: bad args");
+   if (deg(a) >= F.n) LogicError("TransMulMod: bad args");
 
    ZZ_pEX t1, t2;
 
@@ -2390,8 +2390,12 @@ static
 void ProjectPowers(vec_ZZ_pE& x, const ZZ_pEX& a, long k, 
                    const ZZ_pEXArgument& H, const ZZ_pEXModulus& F)
 {
-   if (k < 0 || NTL_OVERFLOW(k, 1, 0) || deg(a) >= F.n) 
-      Error("ProjectPowers: bad args");
+   if (k < 0 || deg(a) >= F.n) 
+      LogicError("ProjectPowers: bad args");
+
+   if (NTL_OVERFLOW(k, 1, 0)) 
+      ResourceError("ProjectPowers: excessive args");
+
 
    long m = H.H.length()-1;
    long l = (k+m-1)/m - 1;
@@ -2420,7 +2424,7 @@ void ProjectPowers(vec_ZZ_pE& x, const ZZ_pEX& a, long k, const ZZ_pEX& h,
                    const ZZ_pEXModulus& F)
 {
    if (k < 0 || deg(a) >= F.n || deg(h) >= F.n)
-      Error("ProjectPowers: bad args");
+      LogicError("ProjectPowers: bad args");
 
    if (k == 0) {
       x.SetLength(0);;
@@ -2520,8 +2524,8 @@ void BerlekampMassey(ZZ_pEX& h, const vec_ZZ_pE& a, long m)
 
 void MinPolySeq(ZZ_pEX& h, const vec_ZZ_pE& a, long m)
 {
-   if (m < 0 || NTL_OVERFLOW(m, 1, 0)) Error("MinPoly: bad args");
-   if (a.length() < 2*m) Error("MinPoly: sequence too short");
+   if (m < 0 || NTL_OVERFLOW(m, 1, 0)) LogicError("MinPoly: bad args");
+   if (a.length() < 2*m) LogicError("MinPoly: sequence too short");
 
    BerlekampMassey(h, a, m);
 }
@@ -2539,7 +2543,7 @@ void DoMinPolyMod(ZZ_pEX& h, const ZZ_pEX& g, const ZZ_pEXModulus& F, long m,
 void ProbMinPolyMod(ZZ_pEX& h, const ZZ_pEX& g, const ZZ_pEXModulus& F, long m)
 {
    long n = F.n;
-   if (m < 1 || m > n) Error("ProbMinPoly: bad args");
+   if (m < 1 || m > n) LogicError("ProbMinPoly: bad args");
 
    ZZ_pEX R;
    random(R, n);
@@ -2556,7 +2560,7 @@ void MinPolyMod(ZZ_pEX& hh, const ZZ_pEX& g, const ZZ_pEXModulus& F, long m)
 {
    ZZ_pEX h, h1;
    long n = F.n;
-   if (m < 1 || m > n) Error("MinPoly: bad args");
+   if (m < 1 || m > n) LogicError("MinPoly: bad args");
 
    /* probabilistically compute min-poly */
 
@@ -2588,7 +2592,7 @@ void MinPolyMod(ZZ_pEX& hh, const ZZ_pEX& g, const ZZ_pEXModulus& F, long m)
 
 void IrredPolyMod(ZZ_pEX& h, const ZZ_pEX& g, const ZZ_pEXModulus& F, long m)
 {
-   if (m < 1 || m > F.n) Error("IrredPoly: bad args");
+   if (m < 1 || m > F.n) LogicError("IrredPoly: bad args");
 
    ZZ_pEX R;
    set(R);
@@ -2706,7 +2710,7 @@ long OptWinSize(long n)
 void PowerMod(ZZ_pEX& h, const ZZ_pEX& g, const ZZ& e, const ZZ_pEXModulus& F)
 // h = g^e mod f using "sliding window" algorithm
 {
-   if (deg(g) >= F.n) Error("PowerMod: bad args");
+   if (deg(g) >= F.n) LogicError("PowerMod: bad args");
 
    if (e == 0) {
       set(h);
@@ -2816,18 +2820,20 @@ void PowerMod(ZZ_pEX& h, const ZZ_pEX& g, const ZZ& e, const ZZ_pEXModulus& F)
 
 void InvMod(ZZ_pEX& x, const ZZ_pEX& a, const ZZ_pEX& f)
 {
-   if (deg(a) >= deg(f) || deg(f) == 0) Error("InvMod: bad args");
+   if (deg(a) >= deg(f) || deg(f) == 0) LogicError("InvMod: bad args");
 
-   ZZ_pEX d, t;
+   ZZ_pEX d, xx, t;
 
-   XGCD(d, x, t, a, f);
+   XGCD(d, xx, t, a, f);
    if (!IsOne(d))
-      Error("ZZ_pEX InvMod: can't compute multiplicative inverse");
+      InvModError("ZZ_pEX InvMod: can't compute multiplicative inverse");
+
+   x = xx;
 }
 
 long InvModStatus(ZZ_pEX& x, const ZZ_pEX& a, const ZZ_pEX& f)
 {
-   if (deg(a) >= deg(f) || deg(f) == 0) Error("InvModStatus: bad args");
+   if (deg(a) >= deg(f) || deg(f) == 0) LogicError("InvModStatus: bad args");
    ZZ_pEX d, t;
 
    XGCD(d, x, t, a, f);
@@ -2843,7 +2849,7 @@ long InvModStatus(ZZ_pEX& x, const ZZ_pEX& a, const ZZ_pEX& f)
 void MulMod(ZZ_pEX& x, const ZZ_pEX& a, const ZZ_pEX& b, const ZZ_pEX& f)
 {
    if (deg(a) >= deg(f) || deg(b) >= deg(f) || deg(f) == 0)
-      Error("MulMod: bad args");
+      LogicError("MulMod: bad args");
 
    ZZ_pEX t;
 
@@ -2853,7 +2859,7 @@ void MulMod(ZZ_pEX& x, const ZZ_pEX& a, const ZZ_pEX& b, const ZZ_pEX& f)
 
 void SqrMod(ZZ_pEX& x, const ZZ_pEX& a, const ZZ_pEX& f)
 {
-   if (deg(a) >= deg(f) || deg(f) == 0) Error("SqrMod: bad args");
+   if (deg(a) >= deg(f) || deg(f) == 0) LogicError("SqrMod: bad args");
 
    ZZ_pEX t;
 
@@ -2864,7 +2870,7 @@ void SqrMod(ZZ_pEX& x, const ZZ_pEX& a, const ZZ_pEX& f)
 
 void PowerXMod(ZZ_pEX& hh, const ZZ& e, const ZZ_pEXModulus& F)
 {
-   if (F.n < 0) Error("PowerXMod: uninitialized modulus");
+   if (F.n < 0) LogicError("PowerXMod: uninitialized modulus");
 
    if (IsZero(e)) {
       set(hh);
@@ -2895,7 +2901,7 @@ void reverse(ZZ_pEX& x, const ZZ_pEX& a, long hi)
 {
    if (hi < 0) { clear(x); return; }
    if (NTL_OVERFLOW(hi, 1, 0))
-      Error("overflow in reverse");
+      ResourceError("overflow in reverse");
 
    if (&x == &a) {
       ZZ_pEX tmp;
@@ -2910,7 +2916,7 @@ void reverse(ZZ_pEX& x, const ZZ_pEX& a, long hi)
 void power(ZZ_pEX& x, const ZZ_pEX& a, long e)
 {
    if (e < 0) {
-      Error("power: negative exponent");
+      ArithmeticError("power: negative exponent");
    }
 
    if (e == 0) {
@@ -2931,7 +2937,7 @@ void power(ZZ_pEX& x, const ZZ_pEX& a, long e)
    }
 
    if (da > (NTL_MAX_LONG-1)/e)
-      Error("overflow in power");
+      ResourceError("overflow in power");
 
    ZZ_pEX res;
    res.SetMaxLength(da*e + 1);
@@ -2970,7 +2976,7 @@ void FastTraceVec(vec_ZZ_pE& S, const ZZ_pEXModulus& f)
 void PlainTraceVec(vec_ZZ_pE& S, const ZZ_pEX& ff)
 {
    if (deg(ff) <= 0)
-      Error("TraceVec: bad args");
+      LogicError("TraceVec: bad args");
 
    ZZ_pEX f;
    f = ff;
@@ -3012,13 +3018,9 @@ void TraceVec(vec_ZZ_pE& S, const ZZ_pEX& f)
 }
 
 static
-void ComputeTraceVec(const ZZ_pEXModulus& F)
+
+void ComputeTraceVec(vec_ZZ_pE& S, const ZZ_pEXModulus& F)
 {
-   vec_ZZ_pE& S = *((vec_ZZ_pE *) &F.tracevec);
-
-   if (S.length() > 0)
-      return;
-
    if (F.method == ZZ_pEX_MOD_PLAIN) {
       PlainTraceVec(S, F.f);
    }
@@ -3032,18 +3034,24 @@ void TraceMod(ZZ_pE& x, const ZZ_pEX& a, const ZZ_pEXModulus& F)
    long n = F.n;
 
    if (deg(a) >= n)
-      Error("trace: bad args");
+      LogicError("trace: bad args");
 
-   if (F.tracevec.length() == 0) 
-      ComputeTraceVec(F);
+   do { // NOTE: thread safe lazy init
+      Lazy<vec_ZZ_pE>::Builder builder(F.tracevec.val());
+      if (!builder()) break;
+      UniquePtr<vec_ZZ_pE> p;
+      p.make();
+      ComputeTraceVec(*p, F);
+      builder.move(p);
+   } while (0);
 
-   InnerProduct(x, a.rep, F.tracevec);
+   InnerProduct(x, a.rep, *F.tracevec.val());
 }
 
 void TraceMod(ZZ_pE& x, const ZZ_pEX& a, const ZZ_pEX& f)
 {
    if (deg(a) >= deg(f) || deg(f) <= 0)
-      Error("trace: bad args");
+      LogicError("trace: bad args");
 
    project(x, TraceVec(f), a);
 }
@@ -3108,7 +3116,7 @@ void resultant(ZZ_pE& rres, const ZZ_pEX& a, const ZZ_pEX& b)
 void NormMod(ZZ_pE& x, const ZZ_pEX& a, const ZZ_pEX& f)
 {
    if (deg(f) <= 0 || deg(a) >= deg(f)) 
-      Error("norm: bad args");
+      LogicError("norm: bad args");
 
    if (IsZero(a)) {
       clear(x);
@@ -3248,7 +3256,7 @@ void PrecomputeProj(vec_ZZ_p& proj, const ZZ_pX& f)
 {
    long n = deg(f);
 
-   if (n <= 0) Error("PrecomputeProj: bad args");
+   if (n <= 0) LogicError("PrecomputeProj: bad args");
 
    if (ConstTerm(f) != 0) {
       proj.SetLength(1);
@@ -3270,8 +3278,11 @@ void ProjectPowersTower(vec_ZZ_p& x, const vec_ZZ_pE& a, long k,
 {
    long n = F.n;
 
-   if (a.length() > n || k < 0 || NTL_OVERFLOW(k, 1, 0)) 
-      Error("ProjectPowers: bad args");
+   if (a.length() > n || k < 0) 
+      LogicError("ProjectPowers: bad args");
+   if (NTL_OVERFLOW(k, 1, 0)) 
+      ResourceError("ProjectPowers: excessive args");
+
 
    long m = H.H.length()-1;
    long l = (k+m-1)/m - 1;
@@ -3307,7 +3318,7 @@ void ProjectPowersTower(vec_ZZ_p& x, const vec_ZZ_pE& a, long k,
                    const vec_ZZ_p& proj)
 
 {
-   if (a.length() > F.n || k < 0) Error("ProjectPowers: bad args");
+   if (a.length() > F.n || k < 0) LogicError("ProjectPowers: bad args");
 
    if (k == 0) {
       x.SetLength(0);
@@ -3339,7 +3350,7 @@ void ProbMinPolyTower(ZZ_pX& h, const ZZ_pEX& g, const ZZ_pEXModulus& F,
 {
    long n = F.n;
    if (m < 1 || m > n*ZZ_pE::degree())
-      Error("MinPoly: bad args");
+      LogicError("MinPoly: bad args");
 
    vec_ZZ_pE R;
    R.SetLength(n);
@@ -3359,7 +3370,7 @@ void ProbMinPolyTower(ZZ_pX& h, const ZZ_pEX& g, const ZZ_pEXModulus& F,
 {
    long n = F.n;
    if (m < 1 || m > n*ZZ_pE::degree())
-      Error("MinPoly: bad args");
+      LogicError("MinPoly: bad args");
 
    vec_ZZ_pE R;
    R.SetLength(n);
@@ -3377,7 +3388,7 @@ void MinPolyTower(ZZ_pX& hh, const ZZ_pEX& g, const ZZ_pEXModulus& F, long m)
    long n = F.n;
 
    if (m < 1 || m > n*ZZ_pE::degree())
-      Error("MinPoly: bad args");
+      LogicError("MinPoly: bad args");
 
    vec_ZZ_p proj;
    PrecomputeProj(proj, ZZ_pE::modulus());
@@ -3417,7 +3428,7 @@ void MinPolyTower(ZZ_pX& hh, const ZZ_pEX& g, const ZZ_pEXModulus& F, long m)
 void IrredPolyTower(ZZ_pX& h, const ZZ_pEX& g, const ZZ_pEXModulus& F, long m)
 {
    if (m < 1 || m > deg(F)*ZZ_pE::degree())
-      Error("IrredPoly: bad args");
+      LogicError("IrredPoly: bad args");
 
    vec_ZZ_pE R;
    R.SetLength(1);
